#ifdef IGNORE_ME
  #define ident "@(#) c_fixed_length_record.cc (/sccs/ri_util/src/s.c_fixed_length_record.cc) 1.2 02/24/99 14:20:45";
#endif

/*
 * file:  c_fixed_length_record.cpp
 *
 * description: implementation of c_fixed_length_record class
 *
 * revision list:
 *  Wed May  6 18:35:31 EDT 1998 KRB - created file
 *
 *
 *
 *
 */

#include <stdinc.h>
#include "c_fixed_length_record.h"

c_fixed_length_record::c_fixed_length_record( ostream& o )
  : fields(7)
  , out(o)
{
  // do nothing...
}


c_fixed_length_record::c_fixed_length_record( const c_fixed_length_record& that )
  : fields(7)
  , out(that.out)
{
  *this = that;
  // we already copied the other hash table, do nothing else...
}


c_fixed_length_record::~c_fixed_length_record(void)
{
  // do nothing, let the hash destroy itself...
}


int c_fixed_length_record::add_field( const char * name )
{
  return fields.add( c_string(name), c_fixed_length_field(c_string(name), c_string(), 0, 0) );
}


int c_fixed_length_record::add_field( const char* name, const char* value )
{
  c_fixed_length_field field;

  if( !fields.find( c_string(name), field ) ) {
    return fields.add( 
        c_string(name),                            // name
        c_fixed_length_field(
        (unsigned int)0,               // begin
        (unsigned int)0,               // end
        c_string(name),  // value
        c_string(value))
    );
  }

  field.set_value( value );

  return fields.add(c_string(name), field);
}


int c_fixed_length_record::add_field( 
  unsigned int begin, 
  unsigned int end, 
  const char* name, 
  const char* value 
) {
  #if defined( _DEBUG )
    INFO(out) << "adding: (" << begin << "," << end << "," << name << ")" << endl;
  #endif
  return fields.add( c_string(name), c_fixed_length_field( begin, end, name, value ) );
}


int c_fixed_length_record::add_field(
  const char* name,
  const char* _begin,
  const char* _end,
  const char* value
) {
  int begin = (unsigned int) atoi(_begin);
  int end   = (unsigned int) atoi(_end);

  return fields.add( c_string(name), c_fixed_length_field( begin, end, name, value ) );
};

int c_fixed_length_record::add_field(const c_fixed_length_field& field )
{
  c_string name;
  field.get_name(name);
  return fields.add( name, field );
}

const c_fixed_length_record& c_fixed_length_record::operator=( 
  const c_fixed_length_record& right ) 
{
  // **b $FileName:$LineNumber
  fields.remove_all();
  c_collection_iter<c_map_pair<c_string,c_fixed_length_field> > *iter;
  iter = right.fields.prepare_iter();
  c_map_pair<c_string,c_fixed_length_field> pair;
  while( right.fields.next( iter, pair ) ) {
    fields.add( pair );
  }
  right.fields.complete_iter(iter);

  return *this;
}

int c_fixed_length_record::operator==( const c_fixed_length_record& right ) const
{
  if( fields.get_count() != right.fields.get_count() ) {
    return FALSE;
  }

  // don't compare the out objects...just the field lists...
  c_collection_iter<c_map_pair<c_string,c_fixed_length_field> > *my_iter, *its_iter;
  c_map_pair<c_string,c_fixed_length_field> my_pair, its_pair;

  my_iter  = fields.prepare_iter();
  its_iter = right.fields.prepare_iter();
  while( fields.next( my_iter, my_pair) && right.fields.next( its_iter, its_pair ) ) {
    if( my_pair.m_key != its_pair.m_key ) {
      return FALSE;
    }
  }
  fields.complete_iter(my_iter);
  right.fields.complete_iter(its_iter);

  return TRUE;
}


int c_fixed_length_record::get_value( const char* field, char* value ) const
{
  value[0] = '\0';

  c_map_pair<c_string,c_fixed_length_field> pair;
  pair.m_key = field;
  c_string tmp;

  if( fields.find( pair ) ) {
    pair.m_value.get_value(tmp);
    strcpy( value, tmp );
    return TRUE;
  }

  return FALSE;
}

int c_fixed_length_record::remove_field( const char* field )
{
  return fields.remove(field);
}


int c_fixed_length_record::get_value( const char* field, c_string& value ) const
{
  value = "";
  c_map_pair<c_string,c_fixed_length_field> pair;
  pair.m_key = field;

  if( fields.find( pair ) ) {
    pair.m_value.get_value(value);
    return TRUE;
  }

  return FALSE;
}


int c_fixed_length_record::get_value( const c_string& field, c_string& value ) const
{
  value = "";
  c_map_pair<c_string,c_fixed_length_field> pair;
  pair.m_key = field;
  if( fields.find( pair ) ) {
    pair.m_value.get_value(value);
    return TRUE;
  }

  return FALSE;
}

int c_fixed_length_record::parse( const char* s )
{
  c_collection_iter<c_map_pair<c_string,c_fixed_length_field> > *iter;
  c_map_pair<c_string,c_fixed_length_field> pair;
  int begin, end;
  c_string line(s), tmp;

  iter = fields.prepare_iter();
  while( fields.next( iter, pair ) ) {
    begin = pair.m_value.get_begin();
    end = pair.m_value.get_end();

    if( !line.mid( begin, end - begin + 1, tmp ) ) {
      // error...
      c_string name;
      pair.m_value.get_name(name);
      WARN(cout) << "Error, length " << line.get_length() 
                 << " of string argument to parse is less " << endl;
      WARN(cout) << "  than the end position(" << end << ") for field: " 
                 << name << endl;
      continue;
    }

    pair.m_value.set_value( tmp );
    
    // re-add the pair so the one in the hashtable get's updated...
    fields.add(pair);
  }
  fields.complete_iter(iter);

  return TRUE;
}

ostream& operator<<( ostream& o, const c_fixed_length_record& record )
{
  // o << "[c_fixed_length_record] num fields: " << record.fields.get_count() << endl;
  // **b $FileName:$LineNumber
  c_collection_iter<c_map_pair<c_string,c_fixed_length_field> > *iter;
  c_map_pair<c_string,c_fixed_length_field> pair;

  INFO(o) << "[c_fixed_length_record] Dump:" << endl;
  INFO(o) << "[c_fixed_length_record] num fields: " << record.fields.get_count() << endl;
  iter = record.fields.prepare_iter();
  while( record.fields.next( iter, pair ) ) {
    o << pair.m_value;
  }
  record.fields.complete_iter(iter);
  return o;
}

void c_fixed_length_record::dump( void ) const
{
  // **b $FileName:$LineNumber
  c_collection_iter<c_map_pair<c_string,c_fixed_length_field> > *iter;
  c_map_pair<c_string,c_fixed_length_field> pair;

  INFO(out) << "[c_fixed_length_record] Dump:" << endl;
  INFO(out) << "[c_fixed_length_record] num fields: " << fields.get_count() << endl;
  iter = fields.prepare_iter();
  while( fields.next( iter, pair ) ) {
    out << pair.m_value;
  }
  fields.complete_iter(iter);
}

void c_fixed_length_record::clear_fields(void)
{
  c_collection_iter<c_map_pair<c_string,c_fixed_length_field> > *iter;
  c_map_pair<c_string,c_fixed_length_field> pair;

  iter = fields.prepare_iter();
  while( fields.next( iter, pair ) ) {
    pair.m_value.set_value(c_string());
    fields.add(pair);
  }
  fields.complete_iter(iter);
}

void c_fixed_length_record::clear(void)
{  
  fields.remove_all();
}

int c_fixed_length_record::get_field_names( c_collection<c_string>& names ) const 
{
  c_collection_iter<c_map_pair<c_string,c_fixed_length_field> > *iter;
  c_map_pair<c_string,c_fixed_length_field> pair;

  iter = fields.prepare_iter();
  while( fields.next( iter, pair ) ) {
    names.add( pair.m_key );
  }
  fields.complete_iter(iter);

  return TRUE;
}

int c_fixed_length_record::get_field( const char* field_name, c_fixed_length_field& field ) const
{
  if( !fields.find( field_name, field ) ) {
    WARN(out) << "field: " << field_name << " not found in record." << endl;
    return FALSE;
  }

  return TRUE;
}

const c_fixed_length_record& 
  c_fixed_length_record::operator +=( const c_fixed_length_record & right ) 
{
  c_collection_iter<c_map_pair<c_string,c_fixed_length_field> > *iter;
  c_map_pair<c_string,c_fixed_length_field> pair;

  iter = right.fields.prepare_iter();
  while( right.fields.next( iter, pair ) ) {
    fields.add(pair);
  }
  right.fields.complete_iter(iter);

  return *this;
}

const c_fixed_length_record operator+( 
  const c_fixed_length_record& left,
  const c_fixed_length_record& right )
{
  c_fixed_length_record tmp(left);
  tmp += right;

  return tmp;
}
