/*
 * file:         c_perl_wrapper.cc
 *
 * description:  implementation of rc_perl_wrapper class
 *
 * revision list:
 *    Tue Jan  5 13:31:39 EST 1999 KRB/CFAN - created the file
 *
 */
#ifdef IGNORE_ME
  #define ident "@(#) c_perl_wrapper.cc (/sccs/ri_util/src/s.c_perl_wrapper.cc) 1.8 02/24/99 14:20:45";
#endif
#include <iostream.h>
#include <fstream.h>
#include <unistd.h>
#include <stdinc.h>
#include <std_init.h>
#include <c_timestamp.h>
#include <c_log_stream.h>

#include <c_perl_wrapper.h>

#include <EXTERN.h>
#include <perl.h>

c_mutex c_perl_wrapper::perl_mutex;     // construct the static mutex

c_perl_wrapper::c_perl_wrapper()
  : perl_interpreter(NULL)              // default this to null
  , perl_error()                        // blank c_string
  , state(UNINITIALIZED)                // invalid state...
{
  ; // do nothing...
}


int c_perl_wrapper::setup_perl_wrapper(int argc, char** argv)
{
  c_mutex_lock(perl_mutex, cout);

  perl_interpreter = (PerlInterpreter*)perl_alloc();
  if( NULL == perl_interpreter ) {
    #if defined( _DEBUG )
      INFO(cout) << "Error, perl_alloc() returned NULL!!!" << endl;
    #endif

    /* why aren't we doing this anymore?
    throw( rc_exception( __FILE__, __LINE__, 0, 
       "Error allocating memory for perl interpreter.") 
    );
    */

    return false;
  }

  perl_construct((PerlInterpreter*)perl_interpreter);
  if( argc && *argv ) {                        // only do this if there are args...
    perl_parse((PerlInterpreter*)perl_interpreter,NULL,argc,argv,NULL);
    perl_run((PerlInterpreter*)perl_interpreter);
    if (SvTRUE(GvSV(errgv))) {
      perl_error = SvPV(GvSV(errgv), na);      // set our internal error buffer

    #if defined( _DEBUG )
        INFO(cout) << "Error: " << SvPV(GvSV(errgv),na) << endl;
      #endif
  
      /* why aren't we doing this anymore?
      throw( rc_exception( __FILE__, __LINE__, 0, 
          "Error in perl_parse and run interpreter.") 
      );
      */

      return FALSE;
    }
  }

  state = INITIALIZED;
  return TRUE;
}

c_perl_wrapper::~c_perl_wrapper()
{
  c_mutex_lock(perl_mutex, cout);   // ok, now we're protected

  if( INITIALIZED == state ) {
    perl_destruct((PerlInterpreter*)perl_interpreter);
    perl_free((PerlInterpreter*)perl_interpreter);
  }
  #if defined( _DEBUG )
  else {
    INFO(cout) << "c_perl_wrapper::~c_perl_wrapper() -- destructing "
               << "UNINITIALIZED object." << endl;
  }
  #endif
}

c_perl_wrapper::perl_rv c_perl_wrapper::run_perl_sub(
  const c_string& sub_name,
  c_list<c_string>& in_args,
  c_list<c_string>& out_args
)
{
  c_mutex_lock(perl_mutex, cout);   // ok, now we're protected

  if( INITIALIZED != state ) {
    #if defined( _DEBUG )
      INFO(cout) << "Error, unable to succeed - never initialized." << endl;
    #endif
    return FAILURE;
  }

  c_string currstr, ret_value;
  int ret_count;
  char instr[1024];
  perl_rv retno;

  // set up perl for the function call...this is the stuff that's not 
  // thread safe...
  dSP;
  ENTER;
  SAVETMPS;
  PUSHMARK(sp);

  // one at a time, iterate through the input arguments, and add them
  // to the argument list for the perl subroutine...

  c_collection_iter<c_string> *iterstr;
  iterstr = in_args.prepare_iter();
  while (in_args.next(iterstr, currstr)) {               // get the next string...
    #if defined(DEBUG)
      INFO(cout) << "CALLING PERL" << currstr << "##" <<endl;
    #endif
    currstr.get_buff(instr);                             // copy it into the buffer
    XPUSHs(sv_2mortal(newSVpv(instr, strlen(instr) )));  // push it onto the arg stack
  }
  in_args.complete_iter(iterstr);
    
  // here we go...
  PUTBACK;
  sub_name.get_buff(instr);                  // get the name of the subroutine
  ret_count = perl_call_pv(instr, G_ARRAY);  // call it
  SPAGAIN;
  if (SvTRUE(GvSV(errgv))) {                 // anything go wrong?
    retno = FAILURE;                         // yup, keep track...
    perl_error = SvPV(GvSV(errgv), na);      // set our internal error buffer
  }

  if (ret_count > 0) {                       // how many args came back?
    retno = ARRAY;                           // we got an array...
    for (int i = 0; i < ret_count; i++) {    // iterate through the arg stack
      ret_value = (char *)POPp;
      out_args.add(ret_value);
      #if defined(DEBUG)
        INFO(cout) << "RETURN FROM PERL" << i << ", value << "<<  ret_value <<endl;
      #endif
    }
  }  
  else {                                     // undef was returned from the sub
    retno = UNDEF;                           // keep track.  This isn't necessarily
  }                                          // an error...

  if (SvTRUE(GvSV(errgv))) {                 // any perl errors occurr?
    retno = FAILURE;                         // should we be doing this?
    perl_error = SvPV(GvSV(errgv), na);      // save off the error
  }

  // clean up any temporaries that got created...
  PUTBACK;
  FREETMPS;
  LEAVE;

  // whew! we're done with the whole messy business...

  #if defined(DEBUG)
    INFO(cout) << "in run_perl_sub ready to return: " << retno << endl;
  #endif

  return retno;
}

int c_perl_wrapper::get_perl_error( c_string& e )
{
  if( INITIALIZED != state ) {
    #if defined( _DEBUG )
      INFO(cout) << "Error, unable to succeed - never initialized." << endl;
    #endif
    return FALSE;
  }

  e = perl_error;
  return TRUE;
}


int c_perl_wrapper::eval_string( char* s )
{
  c_mutex_lock(perl_mutex, cout);   // ok, now we're protected

  if( INITIALIZED != state ) {
    #if defined( _DEBUG )
      INFO(cout) << "Error, unable to succeed - never initialized." << endl;
    #endif
    return FALSE;
  }

  perl_eval_pv(s,TRUE);

  if( SvTRUE(GvSV(errgv)) ) {                // any perl errors occurr?
    perl_error = SvPV(GvSV(errgv), na);      // save off the error
    #if defined( _DEBUG ) 
      INFO(cout) << "Error evaluating perl code: \"" << s << "\"" << endl;
    #endif
    return FALSE;
  }

  return TRUE;
}

int c_perl_wrapper::get_scalar( char* var, c_string& val )
{
  c_mutex_lock(perl_mutex, cout);   // ok, now we're protected

  if( INITIALIZED != state ) {
    #if defined( _DEBUG )
      INFO(cout) << "Error, unable to succeed - never initialized." << endl;
    #endif
    return FALSE;
  }

  val = SvPV(perl_get_sv(var,FALSE),na);

  if( SvTRUE(GvSV(errgv)) ) {                // any perl errors occurr?
    perl_error = SvPV(GvSV(errgv), na);      // save off the error
    #if defined( _DEBUG ) 
      INFO(cout) << "Error setting var: \"" << var << "\"" << endl;
    #endif
    return FALSE;
  }

  return TRUE;
}

