#ifdef IGNORE_ME
  #define ident "@(#) c_socket.cc (/sccs/ri_util/src/s.c_socket.cc) 1.11 02/24/99 14:20:45";
#endif

/*
 * file: c_socket.cc
 *
 * description:  implementation of the socket object
 *
 *
 * revision list:
 *   Tue Jun 23 09:53:07 EDT 1998 KRB - created
 *   Fri Jul 24 09:43:05 EDT 1998 KRB/YN - major changes to ::connect() based 
 *    on code from the eres module: egn_io.c.  Also made a fix and some other 
 *    changes to ::name_resolve().
 *   Thu Sep 10 08:37:42 EDT 1998 KRB - began implementation of the getpeeraddr(),
 *    and getlocaladdr() functions...
 *   Fri Sep 11 11:23:42 EDT 1998 KRB - improved efficiency of recv(), and 
 *    added state checks to the getpeeraddr(), and getlocaladdr() routines...
 *
 *
 */
#include <iostream.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <arpa/inet.h> 
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <c_socket.h>
#include <c_exception.h>
#include <c_timestamp.h>
typedef size_t socklen_t;

const int c_tcp_socket::default_backlog = 5;

#if defined _HP_UX
  #define SOCKADDR_CAST 
	#define socklen_t int
#else
  #define SOCKADDR_CAST (struct sockaddr*)
#endif

c_tcp_socket::c_tcp_socket( ostream& o )
  : family(AF_INET)
  , type(SOCK_STREAM)
  , protocol(IPPROTO_IP)
  , socket_fd(-1)
  , out(o)
  , state(INVALID)
{
  memset( &this_addr, 0, sizeof( struct sockaddr_in ) );
  memset( &peer_addr, 0, sizeof( struct sockaddr_in ) );

  create_socket();
  if( state != CREATED ) {
    #if defined( _DEBUG )
      INFO(out) << "Erorr, create_socket() failed." << endl;
    #endif
    throw( c_exception("Error creating Socket", __FILE__, __LINE__) );
  }
}

c_tcp_socket::c_tcp_socket( const c_tcp_socket& that ) 
  : family(that.family)
  , type(that.type)
  , protocol(that.protocol)
  , socket_fd(that.socket_fd)
  , out(that.out)
  , state(INVALID)
{
  memcpy( &this_addr, &that.this_addr, sizeof( struct sockaddr_in ) );
  memcpy( &peer_addr, &that.peer_addr, sizeof( struct sockaddr_in ) );
 
  create_socket();
  if( state != CREATED ) {
    #if defined( _DEBUG )
      INFO(out) << "Erorr, create_socket() failed." << endl;
    #endif
    throw( c_exception("Error creating Socket", __FILE__, __LINE__) );
  }
}

c_tcp_socket::~c_tcp_socket( void ) 
{
  close();
}


int c_tcp_socket::create_socket( void )
{
  if( state != INVALID ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, state is not INVALID, unable to create socket." << endl;
    #endif
    return FALSE;
  }

  socket_fd = socket(family,type,protocol);

  if( -1 != socket_fd ) {
    state = CREATED;
  }
  else {
    #if defined( _DEBUG )
      INFO(out) << "Error creating socket.  Return value: " << socket_fd << endl;
    #endif
    state = INVALID;
  }

  return( state );
}


int c_tcp_socket::bind( unsigned short port )
{
  if( state != CREATED ) {
    #if defined( _DEBUG )
      INFO(out) << "Error binding socket, state is not CREATED." << endl;
    #endif

    return FALSE;
  }

  this_addr.sin_port = htons(port);
  this_addr.sin_family = family;
  this_addr.sin_addr.s_addr = htonl( INADDR_ANY );

  if( -1 == ::bind( socket_fd, SOCKADDR_CAST(&this_addr), sizeof( struct sockaddr_in ) ) ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, bind returned -1.  Unable to bind socket." << endl;
    #endif
    return FALSE;
  }

  state = BOUND;

  return TRUE;
}

int c_tcp_socket::connect( const char* addr, unsigned short port )
{
  if( state != CREATED ) {
    #if defined( _DEBUG )
      INFO(out) << "Erorr connecting, state is not CREATED." << endl;
    #endif
    return FALSE;
  }


  peer_addr.sin_family = AF_INET;
  peer_addr.sin_addr.s_addr = name_resolve(addr);

  if( -1 == peer_addr.sin_addr.s_addr ) {
    #if defined( _DEBUG )
      ERROR(out) << "Error resolving ip address from: " << addr << endl;
    #endif
    return FALSE;
  }

  #if defined( _DEBUG )
    INFO(out) << "Name resolved: " << addr << " is: " 
        << peer_addr.sin_addr.s_addr << " port: " << port << endl;
  #endif
  

  peer_addr.sin_port = htons(port);

  int rv = ::connect( socket_fd, SOCKADDR_CAST(&peer_addr), sizeof(struct sockaddr_in) );

  #if defined( _DEBUG ) 
    INFO(out) << "::connect() returned: " << rv << endl;
  #endif

  if( 0 != rv ) {
    #if defined( _DEBUG )
      INFO(out) << "Error connecting to: " << addr << ":" << port << endl;
    #endif

    return FALSE;
  }

  state = CONNECTED;

  return TRUE;
}


int c_tcp_socket::listen( int backlog )
{
  if( state != BOUND ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, can't listen, state isn't BOUND." << endl;
    #endif
    return FALSE;
  }

  if( -1 == ::listen( socket_fd, backlog ) ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, listen returned -1.  Unable to listen on socket." << endl;
    #endif
    return FALSE;
  }

  state = LISTENING;

  return TRUE;
}


int c_tcp_socket::accept( c_tcp_socket& client ) 
{
  if( state != LISTENING ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, can't accept, state isn't LISTENING." << endl;
    #endif
    return FALSE;
  }

#if defined(_HP_UX) || defined(_SUN_OS)
  int addrlen = sizeof( struct sockaddr_in );
#else
  socklen_t addrlen = sizeof( struct sockaddr_in );
#endif

  memset( &client.peer_addr, 0, sizeof( struct sockaddr_in ) );
  client.socket_fd = ::accept( socket_fd, SOCKADDR_CAST(&client.peer_addr), &addrlen );

  if( -1 == client.socket_fd ) {
    return FALSE;
  }

  client.state = CONNECTED;

  return TRUE;
}

int c_tcp_socket::close( void )
{
  if( -1 != socket_fd ) {
    ::close(socket_fd);
    socket_fd = -1;
    state = INVALID;
    return TRUE;
  }

  return FALSE;
}

unsigned long c_tcp_socket::name_resolve( const char* host ) const
{
  struct in_addr addr;
  struct hostent *host_ent;

  #if defined(_SUN_OS)
     #define INADDR_NONE -1
  #endif
  if( (addr.s_addr = inet_addr(host)) == INADDR_NONE ) {
    if( !(host_ent = gethostbyname(host)) ) {
      #if defined( _DEBUG )
  ERROR(out) << "Error resolving ip address for: " << host << endl;
      #endif
      return (unsigned long)-1;
    }

    memcpy(&addr.s_addr, host_ent->h_addr, host_ent->h_length);
  }

  return addr.s_addr;
}



int c_tcp_socket::send( const char* buff, int& bytes )
{
  if( state != CONNECTED ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, can't send, state is not CONNECTED." << endl;
    #endif
    return FALSE;
  }

  // try to send the data...
  int rv = ::send( socket_fd, buff, bytes, 0 );

  return rv == bytes;
}

int c_tcp_socket::send( const char* s )
{
  int len = strlen(s);
  send( s, len );
  return len;
}

int c_tcp_socket::recv( char* buff, int& bytes )
{
  if( state != CONNECTED ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, can't recv, state is not CONNECTED." << endl;
    #endif
    return FALSE;
  }

  return ::recv( socket_fd, buff, bytes, 0 );
}

int c_tcp_socket::get_line( char* buff, int& bytes, const char delim )
{
  if( state != CONNECTED ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, can't recv, state is not CONNECTED." << endl;
    #endif
    return FALSE;
  }

  char ch;
  int rv = 0, num = 0;
  while( rv < bytes ) {
    rv = ::recv( socket_fd, &ch, 1, 0 );
    if( !rv ) {
      break;
    }

    if( ch == delim ) {
      buff[num] = '\0';
      break;
    }

    buff[num] = ch;
    buff[num+1] = '\0';

    num++;
  }

  return num;
}

int c_tcp_socket::get_line( char* buff, int& bytes, const char delim, const char eof )
{
  if( state != CONNECTED ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, can't recv, state is not CONNECTED." << endl;
    #endif
    return FALSE;
  }

  char ch;
  int rv = 0, num = 0;
  int found_eof = FALSE;
  while( num < bytes ) {
    rv = ::recv( socket_fd, &ch, 1, 0 );
    if( !rv ) {
      break;
    }

		/*
    if( ch == eof ) {
      buff[num] = '\0';
      found_eof = TRUE;
      break;
    }
		*/

    if( ch == delim ) {
      buff[num] = '\0'; break; }

		// this may not be true for delim != '\n'...
    if( ch == '\r' ) { continue; }

    buff[num] = ch;
    buff[num+1] = '\0';

    num++;
  }

	/*
  if( found_eof ) {
    close();
    return 0;
  }
	*/

  return num;
}


int c_tcp_socket::data_waiting(void)
{
  if( state != CONNECTED ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, can't test data_waiting, state is not CONNECTED." << endl;
    #endif
    return FALSE;
  }

  int bytes = 0;

  int fds[1];
  fd_set readfds;
  int i, n, maxfd = 1;
  struct timeval tv;

  FD_ZERO(&readfds);
  FD_SET(fds[1],&readfds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;

  n = select( maxfd, &readfds, (fd_set*)0, (fd_set*)0, &tv );
  switch(n) {
    case -1:
      // error
      ERROR(out) << "Error performing select on socketfd..." << endl;
      return FALSE;
      break;
    case 0:
      #if defined( _DEBUG )
        INFO(out) << "select timed out, this is probably an error..." << endl;
      #endif
      break;
    default:
      #if defined( _DEBUG )
        INFO(out) << "select succeeded..." << endl;
      #endif
      // do the ioctl(2) call...
      break;
  }

  return bytes;
}


int c_tcp_socket::getsockaddr( char* buff, int len )
{
  if( state != CONNECTED ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, can't getsockaddr, state is not CONNECTED." << endl;
    #endif
    return FALSE;
  }

#if defined(_SUN_OS)
  int addrlen = sizeof( struct sockaddr_in );
#else
  socklen_t addrlen = sizeof( struct sockaddr_in );
#endif
  getsockname( socket_fd, SOCKADDR_CAST(&this_addr), &addrlen );

  char* p = inet_ntoa(this_addr.sin_addr);
  if( NULL == p || strlen(p) > len ) {
    return FALSE;
  }

  strcpy(buff,p);

  INFO(out) << "sockaddr: " << buff << endl;

  return TRUE;
}


int c_tcp_socket::getpeeraddr( char* buff, int len )
{
  if( state != CONNECTED ) {
    #if defined( _DEBUG )
      INFO(out) << "Error, can't getpeeraddr, state is not CONNECTED." << endl;
    #endif
    return FALSE;
  }

#if defined(_SUN_OS)
  int addrlen = sizeof( struct sockaddr_in );
#else
  socklen_t addrlen = sizeof( struct sockaddr_in );
#endif
  getpeername( socket_fd, SOCKADDR_CAST(&peer_addr), &addrlen );

  char *p = inet_ntoa(peer_addr.sin_addr);
  if( NULL == p || strlen(p) > len ) {
    return FALSE;
  }

  strcpy(buff,p);

  INFO(out) << "peeraddr: " << buff << endl;

  return TRUE;
}

