#ifdef IGNORE_ME
  #define ident "@(#) c_string.cc (/sccs/ri_util/src/s.c_string.cc) 1.6 02/24/99 14:20:45";
#endif

/*  c_string.cpp
  Copyright 1997 Kyle R. Burton

c_string.cpp is free software; you can redistribute it and/or modify
it under the tearms of the GNU General Public License as published by 
the Free Software Foundation; either version 2, or (at your option)
any later version.

c_string.cpp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with c_string.cpp; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/


/////////////////////////////////////////////////////////////////////
// file:  c_string.cpp
//
// description:  implementation of c_string class.
//
// creatd:  Sat 10/25/1997 14:34:17.26 Kyle R. Burton
//
// revision list:
//  10/31/97 KRB - fixed left(), right(), and mid() so that when
//    a len of -1 is passed, the whole string
//  10/31/97 KRB - fixed find(ch) - was returning a valid 
//    position even when ch wasn't found...
//  10/31/97 KRB - fixed left(), right(), and mid() so they
//    can be used on *this - i.e. str.left( 5, str ); now does
//    what you would expect it to.
//  11/21/97 KRB - fixed a problem with copy where you couldn't copy
//    a c_string that was '\0'.
// Wed Dec 24 09:27:43 1997 KRB - changed:
//    c_string::operator const char*(void)
//  to:
//    c_string::operator const char*(void) const
//    
//
#if defined( _DEBUG )
  #include <stdio.h>
#endif

#include <string.h>
#include <stdio.h>    // for sprintf
#include <ctype.h>    // for toupper() tolower()
#include "c_string.h"
#include "c_exception.h"

c_string::c_string( const char* buff ) 
  : m_buff( NULL )
  , m_len(0)
{
  m_len = 1;
  m_buff = new char[m_len];
  m_buff[0] = '\0';
  copy( buff );
}

c_string::c_string( char ch, unsigned int rpt ) 
  : m_buff( NULL )
  , m_len(0)
{
  m_len = rpt + 1;
  m_buff = new char[m_len];
  for( unsigned int i = 0; i < rpt; i++ ) {
    m_buff[i] = ch;
  }

  m_buff[rpt] = '\0';
}

c_string::c_string( const c_string& that ) 
  : m_buff( NULL )
  , m_len(0)
{
  m_len = 1;
  m_buff = new char[m_len];
  m_buff[0] = '\0';
  (*this) = that;
}

c_string::~c_string() 
{
  delete [] m_buff;
}

void c_string::empty( void ) 
{
  m_buff[0] = '\0';
}

const c_string& c_string::operator=( const c_string& right ) 
{
  copy( right.m_buff );
  return(*this);
}

const c_string& c_string::operator=( const char* buff ) 
{
  copy( buff );
  return( *this );
}

const c_string& c_string::operator=( char ch ) 
{
  if( m_len < 2 ) {
    delete [] m_buff;
    m_len = 2;
    m_buff = new char[m_len];
  }
  m_buff[0] = ch;
  m_buff[1] = '\0';

  return( *this );
}

int c_string::operator!=( const c_string& right ) const
{
  return( !(*this == right) );
}

int c_string::operator!=( const char* buff ) const
{
  return( !(*this == buff ) );
}

int c_string::operator!=( char ch ) const
{
  return( !(*this == ch ) );
}

int c_string::operator<( const c_string& right ) const
{
  return( -1 == c_string::strcmp( m_buff, right.m_buff ) );
}

int c_string::operator<( const char* buff ) const
{
  return( -1 == c_string::strcmp( m_buff, buff ) );
}

int c_string::operator<( char ch ) const
{
  if( SAFE_STRLEN( m_buff ) != 1 ) {
    return( FALSE );
  }

  return( m_buff[0] == ch );
}

int c_string::operator<=( const c_string& right ) const
{
  return( (*this == right) || (*this < right) );
}

int c_string::operator<=( const char* buff ) const
{
  return( (*this == buff) || (*this < buff) );
}

int c_string::operator<=( char ch ) const
{
  return( (*this == ch) || (*this < ch) );
}

int c_string::operator>( const c_string& right ) const
{
  return( 1 == c_string::strcmp( m_buff, right.m_buff ) );
}

int c_string::operator>( const char* buff ) const
{
    return( 1 == c_string::strcmp( m_buff, buff ) );
}

int c_string::operator>( char ch ) const
{
  if( SAFE_STRLEN( m_buff ) != 1 ) {
    return( FALSE );
  }

  return( m_buff[0] == ch );
}

int c_string::operator>=( const c_string& right ) const
{
  return( (*this == right) || (*this > right) );
}

int c_string::operator>=( const char* buff ) const
{
  return( (*this == buff) || (*this > buff) );
}

int c_string::operator>=( char ch ) const
{
  return( (*this == ch) || (*this > ch) );
}

int c_string::operator!( void ) const
{
  return( 0 == SAFE_STRLEN( m_buff ) );
}

c_string operator+( const c_string& left, const c_string& right ) 
{
  c_string temp = left;
  temp += right;

  return( temp );
}

const c_string& c_string::operator+=( const c_string& right ) 
{
  int new_len = SAFE_STRLEN(m_buff) + SAFE_STRLEN(right.m_buff) +1;
  if( new_len > m_len ) {
    m_len = new_len;
    char *p = new char[m_len];
    strcpy(p,m_buff);
    delete [] m_buff;
    m_buff = p;
  }

  strcat( m_buff, right.m_buff );

  return *this;
}

const c_string& c_string::operator+=( char ch ) 
{
  char buff[2]; 
  buff[0] = ch;  buff[1] = '\0';
  (*this) += buff;
  return( *this );
}


const c_string& c_string::operator+=( const char* buff ) 
{
  int new_len = SAFE_STRLEN(m_buff) + SAFE_STRLEN(buff) + 1;
  if( new_len > m_len ) {
    m_len = new_len;
    char *p = new char[m_len];
    strcpy(p,m_buff);
    delete [] m_buff;
    m_buff = p;
  }

  strcat(m_buff,buff);

  return(*this);
}

int c_string::operator==( const c_string& right ) const
{
  return( 0 == c_string::strcmp( m_buff, right.m_buff ) );
}

int c_string::operator==( const char* buff )  const
{
  return( 0 == c_string::strcmp( m_buff, buff ) );
}


int c_string::operator==( char ch )  const
{
  if( SAFE_STRLEN( m_buff ) != 1 ) {
    return( FALSE );
  }

  return( m_buff[0] == ch );
}

int c_string::get_length( void )  const
{
  return( ( (char*)NULL == m_buff ) ? 0 : SAFE_STRLEN( m_buff ) );
}

void c_string::get_buff( char *buff )  const
{
  if( m_buff[0] == '\0' ) {
    buff[0] = '\0';
  }

  if( buff != m_buff ) { // prevent strcpy on self 
    strcpy( buff, m_buff );
  }
}

void c_string::set_buff( char* buff ) 
{
  copy( buff );
}

int c_string::find( char ch )  const
{
  char *p = m_buff;
  int pos = -1;

  while( *p != '\0' ) {
    if( *p == ch ) {
      pos = p - m_buff;
      break;
    }
    p++;
  }

  return( pos );
}

int c_string::find( const char* str )  const
{
  char *p = m_buff;
  int pos = -1;
  while( *p != '\0' ) {
    if( 0 == strncmp( p, str, SAFE_STRLEN( str ) ) ) {
      pos = p - m_buff;
      break;
    }
    p++;
  }

  return( pos );
}

int c_string::rfind( char ch )  const
{
  char *p = m_buff + SAFE_STRLEN( m_buff );
  int pos = -1;
  while( p != m_buff ) {
    if( *p == ch ) {
      pos = p - m_buff;
      break;
    }
    p--;
  }

  return( pos );
}

int c_string::rfind( const char* str )  const
{
  char *p = m_buff + SAFE_STRLEN( m_buff );
  int pos = -1;
  while( p != m_buff ) {
    //if( 0 == strncmp(m_buff , str, SAFE_STRLEN( str ) ) ) {
    if( 0 == strncmp(p , str, SAFE_STRLEN( str ) ) ) {
      pos = p - m_buff;
      break;
    }
    p--;
  }

  return( pos );
}

char& c_string::operator[]( int index ) 
{
  if( index < 0 || (unsigned int)index > SAFE_STRLEN( m_buff ) ) {
    throw( c_exception::New( "invalid index", __FILE__, __LINE__ ) );
  }

  return( m_buff[index] );
}

int c_string::left( int len, c_string& lft )  const
{
  if( len < 0 ) {
    lft = m_buff;
    return( TRUE );
  }

  len = (unsigned int)len > SAFE_STRLEN( m_buff ) ? SAFE_STRLEN( m_buff ) : len;

  lft = m_buff;
  lft.m_buff[len] = '\0';

  return( TRUE );
}

int c_string::right( int len, c_string& rt )  const
{
  if( len < 0 ) {
    rt = m_buff;
    return( TRUE );
  }

  len = (unsigned int)len > SAFE_STRLEN( m_buff ) ? SAFE_STRLEN( m_buff ) : len;

  rt = m_buff;
  char* p1 = rt.m_buff, *p2 = p1 + SAFE_STRLEN(p1) - len;

  do {
    *p1++ = *p2++;
  } while(*p2);

  *p1 = '\0';

  return( TRUE );
}

int c_string::mid( int start, int len, c_string& md )  const
{
  if( start < 0 || (unsigned int)start > SAFE_STRLEN( m_buff ) ) {
    return( FALSE );
  }

  md = m_buff;
  char* p1 = md.m_buff;        /* point at the beginning */
  char* p2 = p1 + start;   /* point at the start point */
  /* 
   * point at the end point, which is, start + len, unless that's past the
   * end, which we then just use the end...
   *
   */
  char* p3 = (p2 + len) > (p1+SAFE_STRLEN(md.m_buff)) ?
             (p1+SAFE_STRLEN(md.m_buff)) : (p2+len);

  do {
    *p1++ = *p2++;
  } while( p2 != p3 );
  *p1 = '\0';

  return( TRUE );
}

c_string::operator const char*(void) const
{
  return( m_buff );
}

void c_string::copy( const char* buff ) 
{
  // prevent self copy...
  if( m_buff == buff ) {
    #if defined( _DEBUG )
      cout << __FILE__ << "(" << __LINE__ << ") "
           << "Error, attempting self-copy..." << endl;
    #endif
    return;
  }

  if( NULL == buff || '\0' == *buff ) {
    *m_buff = '\0';
    return;
  }

  int new_len = SAFE_STRLEN(buff) +1;

  if( new_len > m_len ) {
    m_len = new_len;
    char *p = new char[new_len];
    delete [] m_buff;
    m_buff = p;
  }

  strcpy(m_buff,buff);
}

const c_string c_string::reverse( void ) const
{
  c_string tmp = m_buff;
  char *p = tmp.m_buff + SAFE_STRLEN( tmp.m_buff ), *p2 = m_buff;
  
  p--;  // leave the null as is

  while( p >= tmp.m_buff && p2 != '\0' ) {
    *p-- = *p2++;
  }

  return( tmp );
}

const c_string c_string::upper( void ) const
{
  c_string tmp = m_buff;

  char *p = tmp.m_buff;

  while( *p != '\0' ) {
    *p = (char)toupper(*p);
    p++;
  }

  return( tmp );
}

const c_string c_string::lower( void ) const
{
  c_string tmp = m_buff;

  char *p = tmp.m_buff;

  while( *p != '\0' ) {
    *p = (char)tolower(*p);
    p++;
  }

  return( tmp );
}

int c_string::eat_white( void )
{
  int rv;

  rv = ltrim();
  rv = rtrim() ? rv : 0;

  return rv;
}

int c_string::ltrim(void) 
{
  char *p1 = m_buff, *p2;

  if( NULL == m_buff || '\0' == *m_buff ) 
    return( FALSE );

  while( *p1 == ' ' || *p1 == '\t' )
    p1++;

  p2 = m_buff;
  while( *p1 != '\0' )
    *p2++ = *p1++;

  *p2 = '\0';


  return( TRUE );
}


  /* this was/is the old rtrim...
   *
   *  p1 = m_buff + SAFE_STRLEN(m_buff) - 1;
   *  while( p1 >= m_buff && (*p1 == ' ' || *p1 == '\t') )
   *    *p1-- = '\0';
   *
   */
int c_string::rtrim( void )
{
  if( NULL == m_buff || '\0' == *m_buff )
    return FALSE;

  char *p = m_buff + SAFE_STRLEN( m_buff ) - 1;

  while( p >= m_buff && (' ' == *p || *p == '\t' ) )
    *p-- = '\0';

  return TRUE;
}

