#ifdef IGNORE_ME
#define ident "@(#) c_tree_mapper.cc (/sccs/ri_util/src/s.c_tree_mapper.cc) 1.9 02/24/99 14:20:45";
#endif

/*
 * file: c_tree_mapper.cc
 *
 * description:  implementation of the tree_mapper object
 *
 *
 * revision list:
 *   Mon Aug 24 09:53:07 EDT 1998 YAN - created
 *   Fri Oct 16 11:44:35 EDT 1998 YAN - fixed a bug in find_node_in_tree -
 *		forgot to return FALSE after the for() loop.
 */

#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <c_list.h>
#include <c_hash.h>
#include <c_string.h>
#include <c_auto_ptr.h>
#include <c_multi_tree.h>
#include <c_timestamp.h>
#include <c_tree_mapper.h>
#include <c_key_val_pair.h>

/*
 * After instantiating a c_tree_mapper object, this method is called to create
 * an unordered multi-node tree.  The user must provide the input list and a
 * template tree
 */
int c_tree_mapper::build_tree(c_list<c_key_val_pair>& input_recs, const c_multi_tree_manipulator<c_string,c_string>& _tmpl_t_hand, const c_multi_tree_manipulator<c_string,c_string>& _unrd_t_hand)
{
	tmpl_t_hand = _tmpl_t_hand;
	unrd_t_hand = _unrd_t_hand;
	list_idx = 0;
	return(process_a_node(input_recs));
}

/*
 * This method is the one that actually builds the unordered tree.  The number
 * of leaves added will be returned.
 */
int c_tree_mapper::process_a_node(c_list<c_key_val_pair>& input_recs)
{
	int match_pos;
	int children;
	int num_leaves_added = 0;
	c_key_val_pair curr_pair;
	c_string k, v;
	static int level = 0;

	level++;
#if defined(_DEBUG)
	INFO(cout) << "[process_a_node] level: " << level << endl;
#endif
	int num_input_recs = input_recs.get_count();
	while (list_idx < num_input_recs)
	{
// get the next input record to process
		input_recs.get_at(list_idx, curr_pair);
// and extract it's key and value
		curr_pair.get_data(k, v);
#if defined(_DEBUG)
		INFO(cout) << "[process_a_node] processing [" << k << "," << v << "] at list_idx " << list_idx << endl;
#endif
		int pos = unrd_t_hand.find_child_pos_by_key(k);
// if key 'k' is one of the children, see if associated value is filled in
		if (pos > -1)
		{
			c_string temp_v;
#if defined(_DEBUG)
			INFO(cout) << "[process_a_node] found key ." << k << ". in unordered tree, let's see what it's value is" << endl;
#endif
			unrd_t_hand.get_child_value(temp_v, pos);
#if defined(_DEBUG)
			INFO(cout) << "[process_a_node] value is ." << temp_v << "." << endl;
#endif
			if (temp_v.get_length())
			{
#if defined(_DEBUG)
				INFO(cout) << "[process_a_node] found key ." << k << ". in unordered tree, with it's value ." << temp_v << ". Means we have to build another subnode" << endl;
#endif
				break;
			}
		}
// see if key 'k' is in a template tree
		match_pos = tmpl_t_hand.find_child_pos_by_value(k);
// if yes, add key, value to the unordered tree...
		if (match_pos > -1) 
		{
			unrd_t_hand.set_child_data(k, v, match_pos);
			list_idx++;
			num_leaves_added++;
		}
		else if (unrd_t_hand.num_children() > 0)
		{
			children = tmpl_t_hand.num_children();
			int num_child_leaves_added = 0;
			int tot_num_child_leaves_added = 0;
// create a matching subnode from template tree in unordered tree
			for ( int i = 0; i < children; ++i )
			{
				if (tmpl_t_hand.childs_num_children(i) == 0)
					continue;
				tmpl_t_hand.down(i);
				tmpl_t_hand.get_data(k, v);
				unrd_t_hand.new_child(k, v);
				unrd_t_hand.last();
				unrd_t_hand.down();
				adopt_children(unrd_t_hand, tmpl_t_hand);
#if defined(_DEBUG)
				c_string table_name;
				tmpl_t_hand.get_value(table_name);
				INFO(cout) << "[process_a_node]: Template's table: " << table_name << endl;
#endif
				num_child_leaves_added = process_a_node(input_recs);
				tot_num_child_leaves_added += num_child_leaves_added;
				unrd_t_hand.up();
				tmpl_t_hand.up();
// if new subnode of the unordered tree has not had any leaves added, delete it
				if (!num_child_leaves_added)
				{
					unrd_t_hand.last();
					unrd_t_hand.delete_child();
				}
			}
//
// if no new subnode was added at all at this level, exit this level because
// the current data we're trying to add does not belong on this level - we have
// to go up
//
			if (!tot_num_child_leaves_added)
			{
#if defined (_DEBUG)
				INFO(cout) << "[process_a_node]: tot_num_child_leaves_added is 0, breaking out" << endl;
#endif
				break;
			}
		}
#if defined(_DEBUG)
	INFO(cout) << "This is list_idx: " << list_idx << endl;
#endif
	}
	level--;
#if defined(_DEBUG)
	INFO(cout) << "Popping out of level " << level << endl;
#endif
	return(num_leaves_added);
}

/*
 * This method will grow new tree branches on the existing frame which came
 * from a template tree
 */
int c_tree_mapper::build_tree(c_list<c_key_val_pair>& input_recs, const c_hash<c_string, c_string>& templ_tree_fields_hash, const c_multi_tree_manipulator<c_string,c_string>& tmpl_t_hand)
{
	int i;
	int pos;
	int children;
	int num_input_recs;
	c_key_val_pair curr_pair;
	c_string k, v;
	c_string node_name;
	c_string temp_v, top_v;
	c_multi_tree_manipulator<c_string,c_string> *handle_ptr;

	list_idx = 0;
	num_input_recs = input_recs.get_count();
	while (list_idx < num_input_recs)
	{
// get the next input record to process
		input_recs.get_at(list_idx, curr_pair);
// and extract it's key and value
		curr_pair.get_data(k, v);
#if defined(_DEBUG)
		INFO(cout) << "[build_tree] processing [" << k << "," << v << "] at list_idx " << list_idx << endl;
#endif
		list_idx++;
//
// Determine node name to which the current field (i.e. leaf) belongs
//
		if (!templ_tree_fields_hash.find(k, node_name))
		{
			WARN(cout) << "Key: " << k << " not found in template tree, skipping [" << k << ", " << v << "] record..." << endl;
			continue;
		}
//
// Find the node in the tree
//
		if (!new_tree_hash.find(node_name, handle_ptr))
		{
			ERROR(cout) << "Node: " << node_name << " does not exists in tree, exiting..." << endl;
			return(FALSE);
		}
//
// Before adding the current leaf to the new tree, do a sanity check
//
		pos = handle_ptr->find_child_pos_by_key(k);
// if key 'k' is not one of the children, we have a major error
		if (pos < 0)
		{
			ERROR(cout) << "Node: " << node_name << " does not contain key: " << k << ", exiting..." << endl;
			return(FALSE);
		}
		handle_ptr->get_child_value(temp_v, pos);
#if defined (_DEBUG)
		INFO(cout) << " Node: " << node_name << " has a key: " << k << " and it's value is: " << temp_v << endl;
#endif
		if (!value_set(temp_v))
		{
			handle_ptr->set_child_value(v, pos);
		}
		else
		{
			c_multi_tree<c_string,c_string> dummy_tree;
			c_multi_tree_manipulator<c_string,c_string> tmpl_tree_to_copy(dummy_tree);
			handle_ptr->get_value(top_v);
//
// Now, we need to locate a node in template tree which matches top_v, so we
// can copy it into the tree we're building
//
			if (!find_node_in_tree(top_v, tmpl_t_hand, tmpl_tree_to_copy))
			{
				ERROR(cout) << "Node: " << top_v << " does not exist in template tree, exiting..." << endl;
				return(FALSE);
			}
//
// Need to duplicate from template tree at top_k node on down
//
			c_multi_tree_manipulator<c_string,c_string> temp_h(*handle_ptr);
			temp_h.up();
			temp_h.new_child();
			temp_h.last();
			temp_h.down();
			temp_h.copy(tmpl_tree_to_copy);
//
// Reset key and value data from template tree representation to unordered tree
// representation
//
			set_tree_keys_from_values(temp_h);
//
// Before adding the current leaf to the new tree, do a sanity check
//
			pos = temp_h.find_child_pos_by_key(k);
// if key 'k' is not one of the children, we have a major error
			if (pos < 0)
			{
				ERROR(cout) << "Node: " << node_name << " does not contain key: " << k << ", exiting..." << endl;
				return(FALSE);
			}
			temp_h.set_child_value(v, pos);
//
// Readjust node pointers in the new tree hash table
//
			create_frame_new_tree(temp_h);
		}
	}
	return(TRUE);
}

/*
 * This method will find a node with top_v in the value field and will set
 * a handle pointing to it
 */
int c_tree_mapper::find_node_in_tree(c_string v_to_find, const c_multi_tree_manipulator<c_string, c_string>& _handle, c_multi_tree_manipulator<c_string, c_string>& find_handle)
{
	int children;
	int i;
	c_string v;
	c_multi_tree_manipulator<c_string,c_string> h(_handle);
//
// Make sure we're pointing at a node and not a leaf
//
	if ((children = h.num_children()) < 1)
		return(FALSE);

	h.get_value(v);
	if (v == v_to_find)
	{
		find_handle = h;
		return(TRUE);
	}
	for (i = 0; i < children; i++)
	{
		h.position(i);
		h.down();
		if (find_node_in_tree(v_to_find, h, find_handle))
			return(TRUE);
		h.up();
	}
	return(FALSE);
}

/*
 * This method will set the key to what is in value and value to blank in the 
 * passed in tree
 */
void c_tree_mapper::set_tree_keys_from_values(const c_multi_tree_manipulator<c_string, c_string>& _handle)
{
	c_multi_tree_manipulator<c_string,c_string> handle(_handle);
	int children;
	c_string v;
	c_string temp_k, temp_v;

	children = handle.num_children();
	if (children < 1)
	{
		handle.get_value(v);
		handle.set_data(v, UNINITIALIZED_VALUE);
		handle.get_data(temp_k, temp_v);
	}
	for (int i = 0; i < children; ++i)
	{
		handle.position(i);
		handle.down();
		set_tree_keys_from_values(handle);
		handle.up();
	}
}

/*
 * This si a boolean method which will return FALSE if value is not initialized
 * (i.e contains UNINITIALIZED_VALUE) and TRUE, otherwise
 */
int c_tree_mapper::value_set(const c_string& _value)
{
	if (_value == UNINITIALIZED_VALUE)
		return(FALSE);
	else
		return(TRUE);
}

/*
 * This method will create a hash table of tree node names (as keys) and
 * handles to those node names (as values)
 */
void c_tree_mapper::create_frame_new_tree(const c_multi_tree_manipulator<c_string,c_string>& tmpl_t_hand)
{
	int children;
	int i;
	int handle_exists;
	c_string v;
	c_multi_tree<c_string,c_string> foo_tree;
	c_multi_tree_manipulator<c_string,c_string> *h_ptr;

	if (children = tmpl_t_hand.num_children())
	{
//
// Get data (i.e. key & value) from a node of a tree
//
		tmpl_t_hand.get_value(v);
//
// Before, we add a new record to the hash table, see if handle pointer part
// of the table is already pointing to something.  If it does, use it, if not,
// create a new one
//
		handle_exists = FALSE;
		if (new_tree_hash.find(v, h_ptr))
		{
			if (h_ptr != NULL)
				handle_exists = TRUE;
		}
		if (!handle_exists)
			h_ptr = new c_multi_tree_manipulator<c_string,c_string>(foo_tree);
		*h_ptr = tmpl_t_hand;
		new_tree_hash.add(v, h_ptr);
#if defined (_DEBUG)
		INFO(cout) << "[create_frame_new_tree] just added (v, ptr)=(" << v << ", " << h_ptr << ") to tree frame hash table" << endl;
#endif

		for (i = 0; i < children; i++)
		{
			h_ptr->next();
			h_ptr->down(i);
			create_frame_new_tree(*h_ptr);
			h_ptr->up();
		}
	}
}

/*
 * This method will return a tree handle pointing to passed in node_name.
 * If node_name is not found, FALSE will be returned.  Otherwise, TRUE will
 * be returned.
 */
int c_tree_mapper::get_handle(const c_string node_name, c_multi_tree_manipulator<c_string,c_string>& return_handle)
{
	c_multi_tree_manipulator<c_string,c_string>* p;

	if (new_tree_hash.find(node_name, p))
	{
		return_handle = *p;
		return TRUE;
	}
	return FALSE;
}

/*
 * This method will return c_multi_tree itself
 */
void c_tree_mapper::get_tree(c_multi_tree<c_string,c_string>& return_tree)
{
	return_tree = dummy_tree;
}

/*
 * After the unordered tree has been build, this method will populate all the
 * join keys, as far as the relational model of the DB is concerned.
 */
void c_tree_mapper::populate_join_keys(const c_multi_tree_manipulator<c_string, c_string>& _handle)
{
	c_multi_tree_manipulator<c_string,c_string> h(_handle);
	c_multi_tree_manipulator<c_string,c_string> temp_h(_handle);
	int children;
	c_string k, v;
	char *tmp_ptr;

	children = h.num_children();
	for ( int i = 0; i < children; ++i )
	{
		if (h.childs_num_children(i) == 0)
		{
				h.get_child_key(k, i);
// if lookup is necessary (i.e. key is made up of table_name.field_name)...
				if ((tmp_ptr = strchr(k, '.')) == NULL)
					continue;
				tmp_ptr++;  // point at one char.  beyond '.'
				temp_h = h;
				temp_h.up();
				int pos = temp_h.find_child_pos_by_key(c_string(tmp_ptr));
				if (pos > -1)
				{
					temp_h.get_child_value(v, pos);
//					h.set_child_value(v, i);
					h.set_child_data(c_string(tmp_ptr), v, i);
#if defined(_DEBUG)
					INFO(cout) << "Setting child " << i << " value to " << v << endl;
#endif
				}
		}
		else
		{
			h.down(i);
#if defined(_DEBUG)
			INFO(cout) << "Recursing for child " << i << endl;
#endif
			populate_join_keys(h);
			h.up();
		}
	}
#if defined(_DEBUG)
	INFO(cout) << "Popping out of populate_join_keys()" << endl;
#endif
}

/*
 * This method will simply print out the contents of the tree to standard
 * output
 */
int c_tree_mapper::print_tree(const c_multi_tree_manipulator<c_string, c_string>& _handle)
{
	c_multi_tree_manipulator<c_string,c_string> handle(_handle);
	static int depth = 0;
	char leader[1024];
	char format[1024];
	char buff[1024];
	c_string key, value;
	int children;

	++depth;

	sprintf(format,"%% %dc", 4*depth);
	sprintf(leader, format, ' ' );
	sprintf(buff,"%08d",depth);
	cout << leader << "depth:          " << buff << endl;;
	handle.get_key(key);
	cout << leader << " key:           " << key << endl;
	handle.get_value(value);
	cout << leader << " value:         " << value << endl;
	children = handle.num_children();

	cout << leader << " num children:  " << children << endl;

	for( int i = 0; i < children; ++i )
	{
		handle.position(i);
		handle.down();
		print_tree(handle);
		handle.up();
	}

	--depth;
	return FALSE;
}

/*
 * This method will copy leaf nodes only at a current level (i.e. not subnodes)
 * from source tree to destination tree, returning the number of leaves copied.
 * IMPORTANT: The value in source tree is turned into key on destination key.
 * value in destination tree is initialized to BLANK.
 */
int c_tree_mapper::adopt_children(const c_multi_tree_manipulator<c_string,c_string>& destination_h, const c_multi_tree_manipulator<c_string,c_string>& source_h)
{
	c_multi_tree_manipulator<c_string,c_string> h1(destination_h);
	c_multi_tree_manipulator<c_string,c_string> h2(source_h);
	int children;
	int num_children_adopted = 0;
	c_string k, v;

	children = h2.num_children();
	for (int i = 0; i < children; ++i)
	{
		if (h2.childs_num_children(i) == 0) // only copy leaves
		{
			h2.get_child_data(k, v, i);
			h1.new_child(v, c_string());
			num_children_adopted++;
		}
	}
	return(num_children_adopted);
}

