#ifdef IGNORE_ME
  #define ident "@(#) date.cc (/sccs/ri_util/src/s.date.cc) 1.3 02/24/99 14:20:45";
#endif

/*  date.cpp
  Copyright 1997 Kyle R. Burton

date.cpp is free software; you can redistribute it and/or modify
it under the tearms of the GNU General Public License as published by 
the Free Software Foundation; either version 2, or (at your option)
any later version.

date.cpp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with date.cpp; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
 * file: date.cpp
 *
 * Description: source module for the date functions
 *
 * Created: 05/27/97 Kyle R. Burton, Ksenia Kaliaguina
 *
 */

#include <stdio.h>
#include "date.h"


int LeapYear( int nYear )
{
  /*  
   *  currently (05/27/97) this test will only will work for
   *  years falling within the 20th and 21st centuries
   *
   */
  if( ( nYear % 4 ) == 0 ) {
    return( TRUE );
  }

  return( FALSE );
}



int ValidDate( int nYear, int nMonth, int nDay )
{
  int nDaysInMonth[12] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

  /* 
   * dates are passed to us with months in the range 1-12, 
   * to match it up with the defines/constants, we need to decrement it
   */
  nMonth--;

  /*
   * to speed things up a bit, we can invalidate this date immediately if 
   * the month falls outside the valid range of months...
   */
  if( nMonth < JAN || nMonth > DEC ) {
    return( FALSE ); 
  }

  /* if it's a leap year, set the number of days to 29 for FEB */
  if( LeapYear( nYear ) ) {
    nDaysInMonth[FEB] = 29;
  }

  /* 
   * if the number of days falls outside the range for a given month, it's
   * an invalid date...
   */
  if( nDay < 1 || nDay > nDaysInMonth[nMonth] ) {
    return( FALSE );
  }

  return( TRUE );
}


// the Oracle internal format stores the dates in a propritary way:
// century and year are stored in excess 100's format i.e. 19th century
// is stored as 119, and '97 is stored as 197
// month and day are in excess 1's notation - they range from 1 to 12/31 inclusive
// Hour, min, and sec are in excess 1's notation - they range from 1 to 60 inclusive
//
// This is off by 1 (or 100 for the year fields) compared to the normal C way of 
// doing business...
//
int DateToChar( char* fmt, char* d, char* target )
{
  char* format = (char*)NULL;
  int year = 0, rval = TRUE;

  while( '\0' != *fmt ) {
    switch( *fmt ) {
      case 'D': 
        switch( ++*fmt ) {
          case 'D': // DD
            ++fmt;
            target += sprintf(target,"%02d",d[DAY_OFFSET]);
            break;
          case '\0': // D
          default:
            target += sprintf(target,"%d",d[DAY_OFFSET]);
            break;
        }
        break;

      case 'H':
        switch( ++*fmt ) {
          case 'H': // HH
            ++fmt;
            target += sprintf(target,"%02d",d[HOUR_OFFSET]);
            break;
          case '\0': // H
          default:
            target += sprintf(target,"%d",d[HOUR_OFFSET]);
            break;
        }
        break;

      case 'M':
        switch( ++*fmt ) {
          case 'I': // MI
            ++fmt;
            target += sprintf(target,"%02d",d[MIN_OFFSET]);
            break;
          case 'i': // Mi
            ++fmt;
            target += sprintf(target,"%d",d[MIN_OFFSET]);
            break;
          case 'M': // MM
            ++fmt;
            target += sprintf(target,"%02d",d[MON_OFFSET]);
            break;
          case '\0': // M
          default:
            target += sprintf(target,"%d",d[MON_OFFSET]);
            break;
        }
        break;

      case 'S':
        switch( ++*fmt ) {
          case 'S': // SS
            ++*fmt;
            target += sprintf(target,"%02d",d[SEC_OFFSET]);
            break;
          case '\0': // S
          default:
            target += sprintf(target,"%d",d[SEC_OFFSET]);
            break;
        }
        break;

      case 'Y':
        switch( ++*fmt ) {
          case 'Y': // YY, YYY, or YYYY
            switch( ++*fmt ) {
              case 'Y': // YYY, or YYYY
          switch( ++*fmt ) {
            case 'Y': // YYYY
              ++fmt;
              format = (char*)"%04d";
              year = d[YEAR_OFFSET];
              break;

            case '\0':
            default:
              format = (char*)"%03d";
              year = d[YEAR_OFFSET] % 1000;
              break;
          }
              case '\0':
              default:
          format = (char*)"%02d";
          year = d[YEAR_OFFSET] % 100;
          break;
            }
            break;

          case '\0':
          default:
            format = (char*)"%d";
            year = d[YEAR_OFFSET] % 10;
            break;
        }
        target += sprintf(target,format,year);
        break;
      
      case 'y':
        format = (char*)"%d";
        switch( ++*fmt ) {
          case 'y': // yy, yyy, or yyyy
            switch( ++*fmt ) {
              case 'y': // yyy, or yyyy
          switch( ++*fmt ) {
            case 'y': // yyyy
              ++fmt;
              year = d[YEAR_OFFSET];
              break;

            case '\0':
            default:
              year = d[YEAR_OFFSET] % 1000;
              break;
          }
              case '\0':
              default:
          year = d[YEAR_OFFSET] % 100;
          break;
            }
            break;

          case '\0':
          default:
            year = d[YEAR_OFFSET] % 10;
            break;
        }
        target += sprintf(target,format,year);
        break;

      default:
        // copy the char to the taget string
        *target = *fmt;
        ++fmt; ++target;
        break;
    }
  }

  return( rval );
}

int CharToDate( char* fmt, char* d, char* target )
{
  int year = 0, rval = TRUE;

// <<REVISIT>> this function isn't yet implemented correctly, this char pointer
// will need to be removed when it is (currently it's a copy of DateToChar( ) )
char* format = (char*)NULL;

  while( '\0' != *fmt ) {
    switch( *fmt ) {
      case 'D': 
        switch( ++*fmt ) {
          case 'D': // DD - extract 2 from the date string...
            target[DAY_OFFSET] = 10 * (*d - '0'); 
            ++d;
            target[DAY_OFFSET] += *d - '0';
            break;

          // this is an issue with all any input that
          // is zero padded: if we get something like:
          // 010197 
          // and take the first char as the date, we won't 
          // eat enough characters...
          case '\0': // D
          default:
            target[DAY_OFFSET] = *d - '\0';
            break;
        }
        break;

      case 'H':
        switch( ++*fmt ) {
          case 'H': // HH - extract 2 from the date string...
            target[HOUR_OFFSET] = 10 * (*d - '0');
            ++d;
            target[HOUR_OFFSET] += *d - '0';
            break;
          case '\0': // H
          default:
            target[HOUR_OFFSET] = *d - '0';
            break;
        }
        break;

      case 'M':
        switch( ++*fmt ) {
          case 'I': // MI
            target[MIN_OFFSET] = 10 * (*d - '0');
            ++d;
            target[MIN_OFFSET] += *d - '0';
            break;
          case 'i': // Mi
            target[MIN_OFFSET] = *d - '0';
            break;
          case 'M': // MM
            target[MON_OFFSET] = 10 * (*d - '0');
            ++d;
            target[MON_OFFSET] += *d - '0';
            break;
          case '\0': // M
          default:
            target[MON_OFFSET] += *d - '0';
            break;
        }
        break;

      case 'S':
        switch( ++*fmt ) {
          case 'S': // SS
            target[SEC_OFFSET] = 10 * (*d - '0');
            ++d;
            target[SEC_OFFSET] = *d - '0';
            break;
          case '\0': // S
          default:
            target[SEC_OFFSET] = *d - '0';
            break;
        }
        break;

      case 'Y':
      case 'y': // same - unless it's zero padded I don't know how to 
        // handle it...
        switch( ++*fmt ) {
          case 'Y': // YY, YYY, or YYYY
            switch( ++*fmt ) {
              case 'Y': // YYY, or YYYY
              case 'y':
          switch( ++*fmt ) {
            case 'Y': // YYYY
            case 'y':
              target[CENTURY_OFFSET] = 10 * (*d - '0');
              ++d;
              target[CENTURY_OFFSET] = *d - '0';
              target[CENTURY_OFFSET] += 100;
              ++d;
              target[YEAR_OFFSET] = 10 * (*d - '0');
              ++d;
              target[YEAR_OFFSET] = *d - '0';
              target[YEAR_OFFSET] += 100;
              break;

            case '\0':
            default:
              target[CENTURY_OFFSET] = *d - '0';
              target[CENTURY_OFFSET] += 100;
              ++d;
              target[YEAR_OFFSET] = 10 * (*d - '0');
              ++d;
              target[YEAR_OFFSET] += *d - '0';
              target[YEAR_OFFSET] += 100;
              break;
          }
              case '\0': // YY
              default:
          target[CENTURY_OFFSET] = 100;
          target[YEAR_OFFSET] = 10 * (*d - '0');
          ++d;
          target[YEAR_OFFSET] += *d - '0';
          target[YEAR_OFFSET] += 100;
          break;
            }
            break;

          case '\0': // Y
          default:
            target[CENTURY_OFFSET] = 100;
            target[YEAR_OFFSET] = 100;
            target[YEAR_OFFSET] = *d - '0';
            break;
        }
        break;
      
      default:
        // copy the char to the taget string
        *target = *fmt;
        ++fmt; ++d;
        break;
    }
  }

  return( rval );
}

