#ifdef IGNORE_ME
  #define ident "@(#) null_fgets.cc (/sccs/ri_util/src/s.null_fgets.cc) 1.2 02/24/99 14:20:45";
#endif

/*  null_fgets.cpp
  Copyright 1997 Kyle R. Burton

null_fgets.cpp is free software; you can redistribute it and/or modify
it under the tearms of the GNU General Public License as published by 
the Free Software Foundation; either version 2, or (at your option)
any later version.

null_fgets.cpp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with null_fgets.cpp; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
 * file: null_fgets.cpp
 *
 *
 * created: 06/02/97 Kyle R. Burton
 *
 * revision list:
 *    06/04/97 KRB - implemented the static buffer method to improve performance 
 *  with the assistance of Dana Camacho
 *    08/18/97 KRB - added the use of ppStatic as the offset into szStatic.  I
 *  chose to add a fucntion argument, in lieu of using a static pointer,
 *  so the function will be re-entrant, and multi-thread safe.  This was added
 *  to improve the performance of the function.  Previously I was shifting the
 *   buffer down when the line was removed, now a pointer is maintained, and the
 *   buffer is only shifted when it is refilled.
 *
 */

#include <stdio.h>
#include <string.h>
#include "stdinc.h"
#include "null_fgets.h"

int overlapped_strcpy( char* szTarget, char* szSource );

/*
 * NAME
 *  null_fgets - get a line of text from a stream, translating NULLs.
 *
 * SYNOPSIS
 *      #include <null_fgets.h>
 *
 *  char *null_fgets(  
 *        char ch, 
 *        unsigned int uSize, 
 *        char *szStatic, 
 *        char **ppStatic, 
 *        char *szTarget, 
 *        FILE *stream 
 *      );
 *
 * DESCRIPTION
 *  null_gets reads  characters  from  stream  into  the array  pointed to by
 *  szTarget, untill EOF is encountered , or untill a newline is encountered.
 *  NULL  characters are translated to ch.   The  buffer  szStatic is used to 
 *  speed up reads - the file will be read in  chunks of uSize at a time with 
 *  the  overflow  stored in the  szStatic buffer.  If 0 is passed for nSize, 
 *  or  NULL  is passed for  szStatic,  then szTarget will be ignored and the 
 *  file will be read 1 character at a time.
 *
 *  NOTE: If the static buffer is being used,  szStatic should be zero length 
 *  (szStatic[0]  should  be  '\0')  the  first time this function is called.  
 *  ppStatic is used with this method,  and will point into szStatic inbetween
 *   function calls.
 *
 */
char* null_fgets
( 
  char ch, 
  unsigned int uSize, 
  char* szStatic, 
  char** ppStatic, 
  char* szTarget, 
  FILE* stream 
)
{
  char cNext, *pTmp = NULL, *pTmp2;
  int nPos = 0, nNumRead, nNumToRead;

  szTarget[0] = '\0';

  /* this portion is tested and working as of 06/03/97 */
  if( 0 == uSize || NULL == szStatic ) {
    do {  /* this is NOT recommended, as getc is *REALLY SLOW* */
      cNext = getc( stream );
      if( EOF == cNext ) {
        szTarget[nPos++] = '\0';
        if( nPos > 1 ) {
          return( szTarget );
        }
        else {
          return( NULL );
        }
      }

      szTarget[nPos++] = (cNext == '\0') ? ch : cNext;
    } while( '\n' != cNext );

    szTarget[nPos] = '\0';
    return( szTarget );
  }
  else { /* *b $FileName:$LineNumber { p *szStatic; } */
    /* 
     * use buffered reads... 
     *
     */

    if( '\0' == *ppStatic[0] ) { /* empty buffer */
      /* 
       * first time in, or the buffer was exhausted last time around,
       * we need to do a read...
       *
       */

       if( feof( stream ) ) {
        return( NULL );
       }

       /* there's nothing to shift down for this condition... */
       *ppStatic = szStatic;
       szStatic[0] = '\0';

       /* fall through to the read section... */
    }
    else {
      /*
       * there's some text in the buffer, is it enough for a line?
       *
       */
      pTmp = strchr( *ppStatic, '\n' );

      if( NULL == pTmp ) { /* some text, but no line in buffer */
        /* 
         * There's no line in the buffer, we need to do another read.
         * Shift down from *ppStatic to szStatic, and 
         * fall through to the read section 
         *
         */

        if( strlen( *ppStatic ) >= uSize ) {
          /* error, the buf is full, and no newline */
          return( NULL );
        }

        if( feof( stream ) ) {
          /* 
           * we hit eof last time through, no more to read, just
           * dump what's left, and return... 
           *
           */
          strcpy( szTarget, *ppStatic );
          *ppStatic = szStatic;
          szStatic[0] = '\0';
          return( szTarget );
        }

        pTmp = szStatic;
        pTmp2 = *ppStatic;

        while( *pTmp2 != '\0' ) {
          *pTmp++ = *pTmp2++;
        }

        *pTmp = '\0';
        *ppStatic = szStatic;

        /* fall through to the read section... */
      }
      else { 
        /* 
         * there is a line in the buffer set ppStatic to the next 
         * offset, and return the line
         *
         */
        if( ( pTmp - *ppStatic ) > uSize ) {
          /* the '\n' is past the end of the buffer? */
          return( NULL );
        }

        *pTmp = '\0';
        strcpy( szTarget, *ppStatic );
        *ppStatic = (pTmp+1);

        return( szTarget );
      }
    }

    /*
     * read section 
     *
     */

    /* *b $FileName:$LineNumber */
    /* find the starting position to fill the buffer */
    *ppStatic = szStatic;
    pTmp = *ppStatic + strlen( *ppStatic );

    nNumToRead = (uSize - 1) - strlen( *ppStatic );
    nNumRead = fread( pTmp, sizeof( char ), nNumToRead, stream );

    if( nNumRead != nNumToRead ) {
      if( !feof( stream ) ) {
        /* read error */
        return( NULL );
      }
    }

    /* replace the nulls with ch */
    pTmp2 = pTmp + nNumRead - 1;

    while( pTmp2 >= szStatic ) {
      if( '\0' == *pTmp2 ) {
        *pTmp2 = ch;
      }

      pTmp2--;
    }

    /* null terminate the buffer */
    pTmp[nNumRead] = '\0';

    /* the buffer is full, get a line out of it */
    return( null_fgets( ch, uSize, szStatic, ppStatic, szTarget, stream ) );
  }

  /* can't get here */
  return( NULL );
}

int overlapped_strcpy( char* szTarget, char* szSource ) 
{
  int i = 0;

  while( *szSource ) {
    ++*szTarget = ++*szSource;
    ++i;
  }

  return( i );
}

